/* ignuit - Educational software for the GNOME, following the Leitner
 * flash-card system.
 *
 * Copyright (C) 2008 Timothy Musson
 *
 * <trmusson@ihug.co.nz>
 * http://homepages.ihug.co.nz/~trmusson/programs.html#ignuit
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <config.h>
#include <gnome.h>

#include "main.h"
#include "card.h"
#include "prefs.h"


static const gchar s_blank[] = "";


struct _Card {
    gchar       *front;         /* Front text. */
    gchar       *back;          /* Back text. */
    Group       group;          /* Difficulty. */
    GDate       *date[N_CARD_DATES];  /* Dates: created, tested, expiry */
    gboolean    expired;        /* Is this card due for testing now? */
    guint       n_tests;        /* Number of times tested. */
    guint       n_known;        /* Number of times scored as known. */
    Category    *category;      /* This card's parent category. */
};


struct _Category {
    gchar       *title;
    GList       *cards;     /* List of all cards in this category. */
};


Card*
card_new (void)
{
    Card *c;

    c = g_new0 (Card, 1);

    c->date[DATE_CREATED] = g_date_new ();
    date_today (c->date[DATE_CREATED]);

    return c;
}


void
card_free (Card *c)
{
    gint i;

    g_free (c->front);
    g_free (c->back);

    for (i = 0; i < N_CARD_DATES; i++) {
        if (c->date[i] != NULL) {
            g_date_free (c->date[i]);
        }
    }

    g_free (c);
}


void
card_list_free (GList *list, gboolean free_cards)
{
    if (list != NULL) {
        if (free_cards) {
            g_list_foreach (list, (GFunc)card_free, NULL);
        }
        g_list_free (list);
    }
}


void
card_set_front (Card *c, const char *s)
{
    g_free (c->front);
    c->front = g_strdup (s);
}


void
card_set_back (Card *c, const char *s)
{
    g_free (c->back);
    c->back = g_strdup (s);
}


const gchar*
card_get_front (Card *c)
{
    return c->front ? c->front : s_blank;
}


const gchar*
card_get_back (Card *c)
{
    return c->back ? c->back : s_blank;
}


void
card_set_group (Card *c, Group g)
{
    if (g < GROUP_NEW || g >= N_GROUPS) {
        g = GROUP_NEW;
    }
    c->group = g;
}


void
card_decrease_test_frequency (Card *c)
{
    /* Mark this card for less frequent testing. */

    if (c->group < GROUP_MAX) {
        c->group++;
    }
}


Group
card_get_group (Card *c)
{
    return c->group;
}


void
card_set_category (Card *c, Category *cat)
{
    c->category = cat;
}


Category*
card_get_category (Card *c)
{
    return c->category;
}


void
card_set_date (Card *c, CardDate cd, guint32 date)
{
    if (c->date[cd] == NULL) {
        c->date[cd] = g_date_new ();
    }
    g_date_set_julian (c->date[cd], date);
}


GDate*
card_get_gdate (Card *c, CardDate cd)
{
    return c->date[cd];
}


guint32
card_get_date (Card *c, CardDate cd)
{
    if (c->date[cd] == NULL) {
        return G_DATE_BAD_JULIAN;
    }
    return g_date_get_julian (c->date[cd]);
}


void
card_set_expired (Card *c, gboolean expired)
{
    c->expired = expired;
}


gboolean
card_get_expired (Card *c)
{
    return c->expired;
}


void
card_set_n_tests (Card *c, guint n)
{
    c->n_tests = n;
}


void
card_inc_n_tests (Card *c)
{
    c->n_tests++;
}


guint
card_get_n_tests (Card *c)
{
    return c->n_tests;
}


void
card_set_n_known (Card *c, guint n)
{
    c->n_known = n;
}


void
card_inc_n_known (Card *c)
{
    c->n_known++;
}


guint
card_get_n_known (Card *c)
{
    return c->n_known;
}


gfloat
card_get_score (Card *c)
{
    return (gfloat)card_get_n_known (c) / (gfloat)card_get_n_tests (c) * 100;
}


void
card_reset_statistics (Card *c)
{
    /* Forget this card's dates and scores. */

    if (c->date[DATE_TESTED] != NULL) {
        g_date_free (c->date[DATE_TESTED]);
        c->date[DATE_TESTED] = NULL;
    }
    if (c->date[DATE_EXPIRY] != NULL) {
        g_date_free (c->date[DATE_EXPIRY]);
        c->date[DATE_EXPIRY] = NULL;
    }
    c->n_tests = 0;
    c->n_known = 0;
    c->expired = FALSE;
    c->group = GROUP_NEW;
}


void
card_mark_tested (Prefs *p, Card *c, gboolean known)
{
    /* Score this card as known or unknown, and update other statistics
     * accordingly.
     */

    GDate today;
    guint32 t;


    /* Record the number of tests. */
    card_inc_n_tests (c);

    /* Record date tested, and reset expiry date. */
    date_today (&today);
    t = g_date_get_julian (&today);
    card_set_date_tested (c, t);
    card_set_date_expiry (c, t);
    card_set_expired (c, FALSE);

    if (known) {

        GDate *expiry;

        /* Record the number of successful tests. */

        card_inc_n_known (c);

        /* Set a new expiry date and mark the card for
         * less frequent testing.
         */

        expiry = card_get_gdate_expiry (c);
        g_date_add_days (expiry, prefs_get_schedule (p, card_get_group (c)));

        card_decrease_test_frequency (c);

        return;
    }


    /* Card was unknown - mark it for frequent testing. */

    card_set_group (c, GROUP_1);
}


void
card_to_top (Card *c)
{
    Category *cat;

    /* Move the card to the start of its category. */

    g_assert (c->category != NULL);

    cat = c->category;
    category_remove_card (cat, c);
    category_prepend_card (cat, c);
}


Category*
category_new (const gchar *title)
{
    Category *cat;

    cat = g_new0 (Category, 1);
    cat->title = title ? g_strdup (title) : g_strdup (_("New Category"));
    return cat;
}


void
category_free (Category *cat, gboolean free_cards)
{
    g_assert (cat != NULL);

    g_free (cat->title);
    card_list_free (cat->cards, free_cards);
    g_free (cat);
}


void
category_set_title (Category *cat, const gchar *title)
{
    g_free (cat->title);
    cat->title = g_strdup (title);
}


const gchar*
category_get_title (Category *cat)
{
    return cat->title ? cat->title : s_blank;
}


void
category_append_card (Category *cat, Card *c)
{
    cat->cards = g_list_append (cat->cards, c);
    c->category = cat;
}


void
category_prepend_card (Category *cat, Card *c)
{
    cat->cards = g_list_prepend (cat->cards, c);
    c->category = cat;
}


void
category_remove_card (Category *cat, Card *c)
{
    g_assert (cat == c->category);

    cat->cards = g_list_remove (cat->cards, c);
    c->category = NULL;
}


GList*
category_get_cards (Category *cat)
{
	g_assert (cat != NULL);
    return cat->cards;
}


guint
category_get_n_cards (Category *cat)
{
    g_assert (cat != NULL);
    return cat->cards ? g_list_length (cat->cards) : 0;
}

